'''
differentiation using finite difference method for forward diff, backward diff and central diff
 Derivatives are given till 4th order derivative however, the solution functions may be differentiated again to get still higher
derivatives
'''


import numpy as np



class FiniteDiff():
    '''
    **Input Parameters:**
    -f: Function whose differention is needed
    -h: step size. Usually small
    NOTE: too small a h can cause the results to diverge

    **Returns:**
    <1> Attributes:
    - The function
    - The step size

    <2> Methods:
    - Differentiation upto 4th derivative using Forward first differences- A function
    - Differentiation upto 4th derivative using backward first differences- A function
    - Differentiation upto 4th derivative using central first differences- A function
    NOTE: The function isn't limited to the 4th order derivative. Higher order derivatives can be found by differentiating the 
          the first order derivatives or higher order derivatives over and over again.

    Functions are provided as output for every function and these can be used to evaluate the derivative at all points.
    '''
    def __init__(self, func, h): #x_pts):
        self.func = func   #Differentiating functoion
        self.h = h  # step size


    # Forward differences
    def Forward_First_Deriv(self):
        def f_first_deriv(x):
            return (self.func(x+self.h) - self.func(x))/self.h
        return f_first_deriv
    
    def Forward_Second_Deriv(self):
        def f_second_deriv(x):
            return (self.func(x+2*self.h) - 2*self.func(x + self.h) + self.func(x))/self.h**2
        return f_second_deriv
        
    def Forward_Third_Deriv(self):
        def f_third_deriv(x):
            return (self.func(x + 3*self.h) - 3*self.func(x + 2*self.h) + 3*self.func(x + self.h) - self.func(x))/self.h**3
        return f_third_deriv

    def Forward_Fourth_Deriv(self):
        def f_fourth_deriv(x):
            return (self.func(x + 4*self.h) - 4*self.func(x+3*self.h) + 6*self.func(x + 2*self.h) - 4*self.func(x + self.h) + self.func(x))/self.h**4
        return f_fourth_deriv

    
    
    # Backward differences
    def Backward_First_Deriv(self):
        def b_first_deriv(x):
            return (self.func(x) - self.func(x - self.h))/self.h
        return b_first_deriv
    
    def Backward_Second_Deriv(self):
        def b_second_deriv(x):
            return (self.func(x) - 2*self.func(x - self.h) + self.func(x - 2*self.h))/self.h**2
        return b_second_deriv
        
    def Backward_Third_Deriv(self):
        def b_third_deriv(x):
            return (self.func(x) - 3*self.func(x - self.h) + 3*self.func(x - 2*self.h) - self.func(x - 3*self.h))/self.h**3
        return b_third_deriv

    def Backward_Fourth_Deriv(self):
        def b_fourth_deriv(x):
            return (self.func(x) - 4*self.func(x - self.h) + 6*self.func(x - 2*self.h) - 4*self.func(x - 3*self.h) + self.func(x - 4*self.h))/self.h**4
        return b_fourth_deriv

    # Central differences
    def Central_First_Deriv(self):
        def central_first_deriv(x):
            return (self.func(x + self.h) - self.func(x - self.h))/(2*self.h)
        return central_first_deriv
    
    def Central_Second_Deriv(self):
        def central_second_deriv(x):
            return (self.func(x + self.h) - 2*self.func(x) + self.func(x - self.h))/self.h**2
        return central_second_deriv
        
    def Central_Third_Deriv(self):
        def central_third_deriv(x):
            return (self.func(x + 2*self.h) - 2*self.func(x + self.h) + 2*self.func(x - self.h) - self.func(x - 2*self.h))/(2*self.h**3)
        return central_third_deriv

    def Central_Fourth_Deriv(self):
        def central_fourth_deriv(x):
            return (self.func(x + 2*self.h) - 4*self.func(x + self.h) + 6*self.func(x) - 4*self.func(x - self.h) + self.func(x - 2*self.h))/self.func**4
        return central_fourth_deriv











if __name__ == '__main__':
    import numpy as np
    import matplotlib.pyplot as plt
    f = lambda x: np.sin(x)
    g = lambda x: np.cos(x)
    x_pts = np.linspace(0, 10, 100000)

    h = .005
    plt.style.use(['science', 'grid', 'notebook'])

    differentiator = FiniteDiff(func = f, h = h)
    fd1 = differentiator.Forward_Fourth_Deriv()
    bc1 = differentiator.Backward_Third_Deriv()
    c1 = differentiator.Central_Third_Deriv()

    plt.plot(x_pts, fd1(x_pts), label = 'forward')
    plt.plot(x_pts, bc1(x_pts), label = 'backward')
    plt.plot(x_pts, c1(x_pts), label = 'central')

    plt.legend()
    plt.show()


from matplotlib import pyplot as plt
import numpy as np
from sympy import re


def NewtonRaphson(f, f_deriv, x0, error=0.001, iterations=1000):
    '''
    The formula for this function is given as
        x_new = x_old - f'(x_old)/f(x_old)

    f : f(x) - The function whose root is needed. x is the independent variable.

    f_deriv : derivative of the given function.

    x0 : initial value for starting the iterations.

    error: maximum difference to tolerate between x_{n+1} and x_n to consider x_n as a root of f.
           By default, it is 0.001

    iterations : No. of iterations for the function. By default, it is 1000.
    '''

    x = x0  # Initial value of parameter

    for i in range(iterations):
        y_x = f(x)  # Value of the function at x
        y_deriv_x = f_deriv(x) # Value of the derivative of the function at x

        if y_deriv_x == 0: # Check if derivative of y at x = 0
            y_deriv_x += 10**(-15)
        x_new = x - (y_x / y_deriv_x)  # Calculate x_new

        if abs(x_new - x) <= error: break  # If x_new - x <= tolerance error, break and return result

        x = x_new  # X gets updated for next iteration

    return print(f'An estimate of the root of the function: {x_new}. \n Number of iterations: {i}')


# Secant method for solving for root if a non linear equation




class SecantMethod():
    '''
    This class returns the root of an equation using the Secant Method

    **Input parameters:**

    - func: f(x) The function whose root is needed
    - guess1 : initial first guess
    - guess2: initial second guess
    
    NOTE: The guesses don't need to be on the opposite sides of the root like bisection method 
    but can be on the same side as well.
    - iterations: No. of iterations to carry for the method
    - tolerance : maximum difference between solutions obtained by consecutuve iterations for that 
    iteration to be considered as a root.
    NOTE: This tolerance should be small for a good root.

    **Returns**
    <1> Upon calling, the class returns the following attributes:
    - guess 1
    - guess 2
    - the function
    - no. of iterations
    - tolerance
    - Number of iterations required for the solution

    <2> Upon calling, The class returns the following methods:
    - The solution upon calling the .solve() method

    '''

    def __init__(self, func, guess1, guess2, iterations, tolerance):
        # Parameters
        self.x1 = guess1
        self.x2 = guess2
        self.func = func
        self.iterations = iterations
        self.tolerance = tolerance
    
    def solve(self):
        for i in range(self.iterations):

            # Checking is denominator is ever 0 and if it is, answer = 0
            numerator = (self.x2 - self.x1)*self.func(self.x2)
            denominator = (self.func(self.x2)-self.func(self.x1))
            if denominator == 0: 
                self.required_iterations = 1
                return 0
            # xn = x2 - (x2-x1)/(f2-f1) according to formula
            x_new = self.x2 - (numerator/denominator)
            err = abs(x_new - self.x2)

            # Checking for error with tolerance level
            if err <= self.tolerance: 
                break
            else:
                # Loop gets updated for next iteration
                self.x1 = self.x2
                self.x2 = x_new
        
        # If error is greater than tolerance even after given number of iterations, function throws warning.
        if err > self.tolerance:
            return 'Warning! Convergence failure'
        
        else:
            self.required_iterations = i # Number of required iterations
            return x_new # Ans






if __name__ == '__main__':
    
    def func1(x):
        return x**2 + 4*x + 4
    def deriv1(x):
        return 2*x + 2
    x = np.linspace(-100, 100, 1000)
    plt.style.use(['science', 'notebook', 'grid'])
    plt.plot(x, func1(x))
    plt.show()
    # Secant method
    root_finder = SecantMethod(func = func1, guess1=-5, guess2 = -3, iterations=20000, tolerance=0.00000001)
    print(f'Secant root = {root_finder.solve()}')
    required_iterations = root_finder.required_iterations
    print(required_iterations)



    # Newton Raphson Method
    sol_newton = NewtonRaphson(f = func1, f_deriv= deriv1, x0 = 1, error = 0.0000000001, iterations=20000)





def Integration_trapezoidal(f, lower_lim, upper_lim, intervals=1000):
    '''
        - f:     A function which needs to be integrated.
                The function should have the forn f(x), where x
                is the dependent variable. If other parameters
                are present they need to be mentioned while giving the function
                in the method.
        - lower_lim : Lower limit of integration - A float.
        - upper_lim : Upper limit of integration: a float
        - Intervals : Number of intervals for which integration needs to be calculated.
                    This is used to calculate the 'h' parameter in the Trapezoidal rule.
                    It is set to 1000 by default.



    **Trapezoidal Rule:**
    Integral = h(1/2[f(Xn) + f(X0)] + f(X1) + f(X2) + f(X3) +......+f(X{n-1}))

    Here: X0 = initial lower limit of integration
          Xn = upper limit of integration
          Xi = X0 + i*h
          h = (upper limit - lower limit)/(no. of intervals we wish to break up the interval in).

    ***Returns:*** The final integral
    '''

    step_size = (upper_lim - lower_lim) / intervals  # The step size for the integratiom, i.e 'h' parameter.

    S = 1 / 2 * (f(lower_lim) + f(upper_lim))  # Integral starts with 1/2(f(X0) + f(Xn))

    for i in range(1, intervals):
        S += f(lower_lim + i * step_size)  # f(xi = x0 + ih) gets appended after every loop

    return step_size * S  # final Integral is returned



def Simpson_one_third(f, lower_lim, upper_lim, intervals=1000):
    '''
    ###Input Parameters:###
    - f:        A function which needs to be integrated.
                The function should have the forn f(x), where x
                is the dependent variable. If other parameters
                are present they need to be mentioned while giving the function in the method.

    - lower_lim :    Lower limit of integration - A float.
    - upper_lim :    Upper limit of integration - A float
    - Intervals : Number of intervals for which integration needs to be calculated.
                    This is used to calculate the 'h' parameter in the Trapezoidal rule.
                    It is set to 1000 by default.

    **NOTE:** The Simpsons rule only works for positive number of intervals. If provided with an odd number of
              interval, the function will prompt to input an even number.


    **Simpson's rule:**
    Integral = h/3([f(X0) + f(Xn)] + 4(f(X1) + f(X2) + f(X3) + .......) + 2(f(X2) + f(X4) + f(X6)+........))

    which is essentially = h/3([f(upper_limit)+ f(lower_limit)] + 4(odd terms sum) + 2(even terms sum))

    **NOTE:** Here, in the even terms sum, f(upper limit), though being 'an even term', should not be used.
    Here: X0 = initial lower limit of integration
          Xn = upper limit of integration
          Xi = X0 + i*h
          h = (upper limit - lower limit)/(no. of intervals we wish to break up the interval in).

    ### Returns:### The integral

    '''

    # Checking if the number of intervals is even or odd. Throw error message if it is NOT even.
    if intervals % 2 == 1:
        raise ValueError('Choose a different interval which is even')

    # The h parameter for the method
    step_size = (upper_lim - lower_lim) / intervals

    odd_terms_sum, even_terms_sum = 0, 0  # Odd terms and even terms

    integral = (f(upper_lim) + f(lower_lim))  # Integral starts with f(upper_lim) + f(lower_lim)

    for i in range(1, intervals):
        if i % 2 == 1:  # n = odd
            odd_terms_sum += f(lower_lim + i * step_size)  # Odd terms get updated
        if i % 2 == 0:  # n = even
            even_terms_sum += f(lower_lim + i * step_size)  # Even terms get updated
    integral = step_size / 3 * (
                integral + 4 * odd_terms_sum + 2 * even_terms_sum)  # Integral gets updated with formula.

    return integral  # Returns just the answer



def Simpson_three_eighth(f, lower_lim, upper_lim, intervals=1000):
    '''
    ###Input Parameters:###
    - f:        A function which needs to be integrated.
                The function should have the forn f(x), where x
                is the dependent variable. If other parameters
                are present they need to be mentioned while giving the function in the method.

    - lower_lim :    Lower limit of integration - A float.
    - upper_lim :    Upper limit of integration - A float
    - Intervals : Number of intervals for which integration needs to be calculated.
                    This is used to calculate the 'h' parameter in the Trapezoidal rule.
                    It is set to 1000 by default.

    **NOTE:** The 3/8 Simpsons rule only works for number of intervals which are integral multiples  of 3.
              If provided with an odd number of intervals, the function will prompt to input an appropriate number..


    **Simpson's 3/8 rule:**
    Integral = 3h/8([f(X0) + f(Xn)] + 3(f(X1) + f(X2) + f(X4) + .......) + 2(f(X3) + f(X6) + f(X9)+........))

    which is essentially = 3h/8([f(upper_limit)+ f(lower_limit)] +  2(terms which are multiples of 3 sum) +
    3(remaining terms sum))

    **NOTE:** Here, in the 'multiple of 3' terms sum, f(upper limit), even if it is  'a multiple of 3 term',
    should not be used.
    Here: X0 = initial lower limit of integration
          Xn = upper limit of integration
          Xi = X0 + i*h
          h = (upper limit - lower limit)/(no. of intervals we wish to break up the interval in).

    ### Returns:### The integral

    '''

    # Checking if the number of intervals is even or odd. Throw error message if it is NOT even.
    if intervals % 3 != 0:
        raise ValueError('Choose a different interval which is a multiple of 3')

    # The h parameter for the method
    step_size = (upper_lim - lower_lim) / intervals

    extra_terms_sum, three_multiple_terms_sum = 0, 0  # Odd terms and even terms

    integral = (f(upper_lim) + f(lower_lim))  # Integral starts with f(upper_lim) + f(lower_lim)

    for i in range(1, intervals):
        if i % 3 == 0:  # n = multiple of 3
            three_multiple_terms_sum += f(lower_lim + i * step_size)  # Odd terms get updated
        else:  # n = not multiple of 3
            extra_terms_sum += f(lower_lim + i * step_size)  # Even terms get updated
    integral = 3 * step_size / 8 * (
                integral + 2 * three_multiple_terms_sum + 3 * extra_terms_sum)  # Integral gets updated with formula.

    return integral  # Returns just the answer


# composite rules of integration

class CompositeIntegration():
    '''
    This class calculates the intergral of a function using Composite Simpson's
    Rule and Composite Trapezoidal rule.
    
    **Input parameters:**
    - Function: The function whose integration is neede
    - lowerlim : Float - The lower limit of integration
    - upperlim : Float - The upper limit of integration
    - intervals : Number of intervals to break up the integral in (2000 by default).
    **Note** A huge number of intervals may seem to make the integral more accurate.
        However, the integral may take up excessive computational power.

    **Returns**
    
    <1> Upon calling, the class resturns the attributes:
    - the function
    - the lower limit
    - the upper limit
    - the number of intervals
    - the 'h' parameter in the formula

    <2>Upon calling, the class resturns the methods:
    -  Answer with Composite Trapezoidal Rule
    - Answer with Composite Simpsons Rule


    '''

    def __init__(self, func, lowerlim, upperlim, intervals=2000):
        self.f = func
        self.lowerlim = lowerlim
        self.upperlim = upperlim
        self.n = intervals
        self.h = float(self.upperlim - self.lowerlim)/self.n
    
    def Comp_Trapezoidal_Rule(self):

        # Initialization
        result = 0.5 * (self.f(self.lowerlim) + self.f(self.upperlim))

        for i in range(1, self.n):
            # X term gets updated
            independent_var = self.lowerlim + i*self.h

            # Solution gets updated
            result += self.f(independent_var)

        return result*self.h

    def Comp_Simpsons_Rule(self):
        # Initialization
        result = self.f(self.upperlim) + self.f(self.lowerlim)

        d = [2 , 4]  # Coefficients for the method
        for i in range(1, self.n):
            new_x = self.lowerlim + i*self.h # X gets updated as x_new = x + ih
            result += d[-1] * self.f(new_x)  # Solution gets updated
            d.reverse()  # List gets reversed
        return result*self.h/3



if __name__ == '__main__':
    import numpy as np
    f = lambda x: x*np.sin(x)
    lowerlim = 0
    upperlim = 10
    intervals = 1000

    # Trapezoidal rule
    trap_ans = Integration_trapezoidal(f=f, lower_lim=lowerlim, upper_lim=upperlim, intervals=intervals)

    # Simpsons one third rule
    simp_one_third_ans = Simpson_one_third(f=f, lower_lim=lowerlim, upper_lim=upperlim, intervals=intervals)

    # Simpsons three eight rule
    simp_three_eight_ans = Simpson_three_eighth(f=f, lower_lim=lowerlim, upper_lim=upperlim, intervals=intervals + 2)

    # Composite rules
    integrator = CompositeIntegration(func=f, lowerlim=lowerlim, upperlim=upperlim, intervals=intervals)

    ## Composite Trapezoidal rule
    comptrap_ans = integrator.Comp_Trapezoidal_Rule()

    ## Composite Simpson's rule
    compsimp_ans = integrator.Comp_Simpsons_Rule()

    print(f'trap_ans = {trap_ans}')
    print(f'simp3/8 = {simp_three_eight_ans}')
    print(f'simp1/3 = {simp_one_third_ans}')
    print(f'comp_simp = {compsimp_ans}')
    print(f'comptrap = {comptrap_ans}')
    


from matplotlib import pyplot as plt
import numpy as np

# Lagrange's Interpolation


class LagrangeInterp():
    '''
    This is a class and hence, an object needs to be created for this class.
    Input parameters:
    - x_list: Array-like. The list of independent points for interpolation.
              Doesn't need to be equally spaced.
    - y_list: Array-like. The list of dependent points for interpolation. 

    The class returns the x_list, y_list and the interpolating function upon calling.
    **Note:** 
    The values in x_list don't need to be equally spaced.
    The points where interpolation is needed lie outside of the given
    range of x_list. However, it is not advisable to do this.


    Example code:
    - import numpy as np
    - x_list = [1,2,3,4] # independent
    - y_list = [1,4,9,16] # dependent
    - interpolator = LagrangeInterp(x_list = x_list, y_list = y_list) # Object
    - func = interpolator.GetFunc() # returns the interpolating function.
    - x_pts = np.linspace(1, 9, 1000) # pts where interpolation is needed
    - y_pts = func(x_pts) # solutions
    '''

    def __init__(self, x_list, y_list):
        self.x_list = x_list  # initial independent variables
        self.y_list = y_list  # initial dependent variables

    def GetFunc(self):
        import numpy as np

        def InterFunc(x):
            func = 0  # Initiation value of the function coefficients
            # x and y values zipped as (x, y)
            pts = list(zip(self.x_list, self.y_list))

            for j in pts:
                x_temp = np.array([x[0] for x in pts if x[0] != j[0]])
                '''choosing the x list 
                 excluding one point at every iteration'''

                # numerator and denominator gets updated to 1 for next loop
                numerator, denominator = 1, 1
                for i in range(len(x_temp)):
                    # Numerator gets updated as (x-x_n)
                    numerator *= (x - x_temp[i])
                    denominator *= j[0] - x_temp[i]
                    '''Denominator gets updated as (x0 - x_n) where x_n = j[0], i.e, the excluded point'''
                func += numerator*j[1]/denominator
                x_temp = self.x_list  # X_temp gets updated back to x_list for next loop
            return func
        return InterFunc  # The interpolating function is returned by GetFunc()


# Newton Forward Interpolation
def NewtonForward(x_list, y_list, x):
    '''
     **Input variables:**
    - x_list : values of the independent parameters as a list
    - y_list : values of the independent parameter as a list
    - x: Array or list: Values of x for which interpolation is needed

     Formula used:
     F(a+uh) = F(a) + u/2! D{F(a)} + u(u-1)/3! D2{F(a)} + ..........
     Here: Dn{F(a)} represents the 'n' th forward difference
          a + uh = value of x at which interpolation is needed.
          NOTE: The h is has to be  constant in this method.

    NOTE: This method can also be done using the numpy function 'np.diff'.
          However, we are not using that to preserve the simplicity of the code.

    Returns : List: The interpolated answer for the given data points of the independent
              variable.
    '''


    # Convert x lists and y lists into numpy arrays for faster computation
    x_list = np.array(x_list)
    y_list = np.array(y_list)

    ''' First we will throw some checks to see
    1) If length of x_list and y_list are same
    2) If the difference of consecutive elements in the x list
    is constant or not. This method only works for a constant difference.
    If the difference is not constant, then function will throw an Error and a comment.
    3) If the type of the input parameters is numeric or not.
    '''

    # Check if length of x_list and y_list are same
    if len(x_list) != len(y_list):
        raise AssertionError('x_list and y_list must be of same length')

    # Check if the difference between consecutive elements of x_list are same or not
    if all(i == np.diff(x_list)[0] for i in np.diff(x_list)) == False:
        raise ArithmeticError('Difference of consecutive elements in x_list is not same')

    # Check if x_list is  list-like or not
    if isinstance(x_list, (list, np.ndarray, tuple))  == False:
        raise TypeError('Please input numeric values for x_list ')

    # Check if y_list is list-like or not
    if isinstance(y_list, (list, np.ndarray, tuple))  == False:
        raise TypeError('Please input numeric values for y_list ')

    # Check if x is list-like or not
    if isinstance(x, (list, np.ndarray, tuple))  == False:
        raise TypeError('Please input numeric values for x')

    ''' Now we modify the x_list and the y_list to work for data 
    when the data is taken from in between the data and not from the beginning of the 
    data. 
    NOTE: This is not recommended. For data from the middle of the data, we should
    use Strirling's formula and for data from the end of the data, we should use
    Newton's backard interpolation method
    '''
    solutions = []  # Solution vector for the interpolated points

    for x_pt in x:

        if isinstance(x_pt, (int, float)) == False:
            raise ValueError('Enter numbers in x')
        # Check if the given point lies within the given x data. Otherwise, throw error
        if (x_pt > max(x_list) or x_pt < min(x_list)):
            raise ValueError('Enter a number within the given x data')

        extra_x_indices = []  # Extra unnecessary elements indices in the x_list
        extra_y_indices = []  # Extra unnecessary elements indices in the y_list
        for i in enumerate(x_list):
            if i[1] < x_pt:
                extra_x_indices.append(i[0])
                extra_y_indices.append(i[0])
        if len(extra_y_indices) and len(extra_x_indices) != 0:
            # Make sure that there is 1 element before x_pt in the table
            # So that a difference may be taken
            extra_y_indices.pop()
            extra_x_indices.pop()
        x_list = np.delete(arr=x_list, obj=extra_x_indices)
        y_list = np.delete(arr=y_list, obj=extra_y_indices)

        diff_table = []  # Forward difference table
        x0 = x_list[0]  # Initial value for forward interpolation

        interp_ans = y_list[0]  # Final answer

        h = x_list[1] - x_list[0]  # This has to be constant

        counter = 1

        coeff = (x_pt - x0) / h
        '''This is the 'u' parameter we see in the formula which takes form u/1!, u(u-1)/2!, u(u-1)(u-2)/3!, .....'''

        for i in range(len(y_list), 1, -1):
            ''' Why do we implement a double loop?
            The double loop is implemented to calculate the consecutive forward differences.
            Every 'n' th iteration calculates the 'n' forward difference.
            
            The range of the first loop is [number of y elements, 2] as is obvious because after every iteration,
            the number of elements in the diff tables get reduced by 1 element.
            
            '''
            for j in range(i - 1):
                '''
                The range of the second loop is (i-1) because of the simple reason that the number of elements in
                the consecutive difference tables are decreasing by 1.
                Example: y = [1,2,3,4,5]  {Length = 5}
                First forward difference = [1, 1, 1, 1]  {Length = 4}
                Second forward difference = [0, 0, 0]  {Length = 3}
                Third forward difference = [0, 0]   {Length = 2}
                Fourth forward difference = [0]    {Length = 1}
                '''
                new_diff = y_list[j + 1] - y_list[j]  # Every diff value is the diff between the '(n+1)' th and 'n' th value
                diff_table.append(new_diff)  # Appending values to the difference table
            interp_ans += coeff * diff_table[0]

            '''We take the first element of diff table and use formula.
            Eg: u * D{f(a)} / 1!,    u * (u-1) * D2{f(a)} / 2! and so on.
            Final answer gets updated as 
            ans = f(a) + uD(f(a)) + u(u-1) D2 (f(a))/2! + u(u-1)u(u-2) D3(f(a))/3! + ....
            '''

            coeff *= (coeff - counter) / (counter + 1)  # Coefficient gets updated as u(u-1)(u-2)....(u-n)/(n+1)!
            counter += 1  # Counter gets updated

            diff_table = []  # Clear the difference table for next iteration

        solutions.append(interp_ans)
    return solutions  # The final interpolated answer gets returned








if __name__ == '__main__':
    def func(x):
        return np.sin(x)

    # xlist doesn't need to be evenly spaced
    x_list = [1, 2, 2.5, 3.7,  4, 4.6, 5, 6, 6.7, 7.2, 8.5, 9]
    y_list = func(x_list)
    interpolator = LagrangeInterp(x_list=x_list, y_list=y_list)

    f = interpolator.GetFunc()  # The interpolatin function

    x_pts = np.linspace(0, 10, 1000)
    y_pts = f(x_pts)
    plt.style.use(['science', 'notebook', 'grid'])
    plt.plot(x_pts, y_pts, label='interpolated function')
    plt.plot(x_list, y_list, 'o', markersize=10)
    plt.legend(loc='best')
    plt.show()



from setuptools import setup, find_packages
 
classifiers = [
       "Development Status :: 1 - Planning",
        "Intended Audience :: Education",
        "Programming Language :: Python :: 3",
        "Operating System :: Unix",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
    ]
 
setup(
  name='pyundergraduate',
  version='0.0.3',
  description='A package for numerical analysis with python',
  long_description=open('README.md').read() + '\n\n' + open('CHANGELOG.txt').read(),
  long_description_content_type="text/markdown",
  url='https://github.com/ScientificArchisman/pyundergraduate',  
  author='Archisman Chakraborti',
  author_email='archismanninja@gmail.com',
  license='MIT', 
  classifiers=classifiers,
  keywords=['python', 'scince', 'numerical analysis', 'physics'], 
  packages=find_packages(),
  install_requires=['numpy', 'scipy', 'matplotlib', 'jupyter', 'scienceplots'] 
)


# Euler function to solve a differential equation of 1st, 2nd and 3rd order numerically using Euler's formula.

from cProfile import label
from matplotlib import pyplot as plt
plt.style.use(['science', 'notebook', 'grid'])

def euler(f, ics, step_size, x1):
    ''' 
    This function solves ODEs upto 3rd Order
    **Input variables:**
    f :

    - If a 1st order differential equation is required to solve,
                  f is differential equation of the form :
                   dy/dx = f(y, x)

    - If a 2nd order ODE is required to  be solved then f = du/dx(u, y, x) where
        u is dy/dx, i.e the first order derivative of the dependent variable with respect to the
        independent variable.

    - If a 3rd order ODE is required to be solved then f = dv/dx(v, u, y, x) where v = y''(x), i.e, the second derivative
        of the dependent variable w.r.t to the independent variable and u = y'(x), i.e the first derivative of the
        dependent variable with respect to the independent variable.

    ics: Initial conditions which should contain atleast 2(including x0, which is compulsory) of:
        - x0 = starting value of independent variable (Example: time)
        - y0 = initial value of dependent variable (Example: position)
        - y'0 = initial value of the first derivative of the dependent variable (Example: velocity)
        - y''0 = initial value of the second derivative of the dependent variable (Example: Acceleration)
        The order should be (independent variable, dependent variable, 1st derivative of dependent variable,
        second derivative of dependent variable).

        Example: ics = - 3rd order ODE : [initial time, initial amplitude, initial velocity, initial acceleration].
                       - 2nd order ODE: [initial time, initial amplitude, initial velocity]
                       - 1st order ODE: [initial time, initial amplitude]

    step_size = value of the 'h' parameter in Euler's method

    x1 = final value of independent variable

    **Returns:**
    - y_deriv_deriv_vals = solution vector of y''(x) if a 3rd order ODE is being solved
    - y_deriv_vals = solution vector of y'(x) if a 3rd order ODE or 2nd order ODE is being solved
    - y_vals = solution vector of dependent variable.
    - x_vals = vector of the independent variable for which the
    dependent variable was calculated'''

    # For solving 1st order ODEs
    if len(ics) == 2:

        x0 = ics[0]  # Initial value of independent variable
        y0 = ics[1]  # Iniial value of dependent variable

        x_vals = [x0]  # Solution vector for the independent variable
        y_vals = [y0]  # Solution vector for the dependent variable
        # Number of iterations based on given step size
        iterations = round((x1 - x0) / step_size)
        '''Number of almost equally spaced points for
         the given step_size and range(x1-x0).
         Iterations are rounded to be able to be used as a range'''

        for i in range(int(iterations)):
            next_approx = y_vals[-1] + step_size * f(y_vals[-1], x_vals[-1])
            '''last element of y vector gets updates as 
                y(x+h) = y(x) + h*y'(x)'''

            y_vals.append(next_approx)
            x_vals.append(x_vals[-1] + step_size)

        return y_vals, x_vals

    # For solving 2nd order ODEs
    if len(ics) == 3:

        x0 = ics[0]  # Initial value of independent parameter
        y0 = ics[1]  # Initial value of dependent parameter
        # Initial value of first derivative of dependent parameter.
        y_deriv0 = ics[2]

        x_vals = [x0]  # Solution vector for independent variable
        y_vals = [y0]  # Solution vector for the dependent variable
        # Solution vector for the first derivative of dependent variable
        y_deriv_vals = [y_deriv0]

        # Number of iterations based on given step size.
        iterations = round((x1 - x0) / step_size)
        '''Number of almost equally spaced points for
             the given step_size and range(x1-x0).
             Iterations are rounded to be able to be used as a range'''

        for i in range(int(iterations)):
            '''Formulas used:
            - x(n+1) = x(n) + h
            - y(n+1) = y(n) + h*(y_deriv(n))
            - y'(n+1) = y'(n) + h*d(y')/dx (y'(n), y(n), x(n))'''

            # New values of dy/dx, y, x
            y_new = y_vals[-1] + step_size * y_deriv_vals[-1]
            y_deriv_new = y_deriv_vals[-1] + step_size * f(y_deriv_vals[-1], y_vals[-1], x_vals[-1])
            x_new = x_vals[-1] + step_size

            # New values get appended to solution vectots
            y_vals.append(y_new)
            y_deriv_vals.append(y_deriv_new)
            x_vals.append(x_new)

        return y_deriv_vals, y_vals, x_vals

    # For solving 3rd order ODEs:

    if len(ics) == 4:
        x0 = ics[0]  # Initial value of independent parameter
        y0 = ics[1]  # Initial value of dependent parameter
        # Initial value of first derivative of dependent parameter
        y_deriv0 = ics[2]
        # Initial value of second derivative of dependent parameter
        y_deriv_deriv0 = ics[3]

        x_vals = [x0]  # Solution vector for independent variable
        y_vals = [y0]  # Solution vector for the dependent variable
        # Solution vector for the first derivative of dependent variable
        y_deriv_vals = [y_deriv0]
        # solution vector for the second derivative of dependent variable
        y_deriv_deriv_vals = [y_deriv_deriv0]

        # Number of iterations based on given step size.
        iterations = round((x1 - x0) / step_size)
        '''Number of almost equally spaced points for
             the given step_size and range(x1-x0).
             Iterations are rounded to be able to be used as a range'''

        # Calculating new values of y''(x), y'(x), y(x), x and making solution vectors

        for i in range(int(iterations)):
            '''Formulas used:
            - x(n+1) = x(n) + h
            - y(n+1) = y(n) + h*(y_deriv(n))
            - y'(n+1) = y'(n) + h*d(y')/dx 
            - y''(n+1) = y''(n) + h*f(y'')/d(x){x, y, y', y''}
            '''

            # new values of dependent variable
            y_new = y_vals[-1] + step_size * y_deriv_vals[-1]
            # new values of y'(x)
            y_deriv_new = y_deriv_deriv_vals[-1] + \
                step_size * y_deriv_deriv_vals[-1]
            y_deriv_deriv_new = y_deriv_deriv_vals[-1] + step_size * f(y_deriv_deriv_vals[-1], y_deriv_vals[-1],
                                                                       y_vals[-1], x_vals[-1])  # new values of y''(x)

            # new values of independent variable
            x_new = x_vals[-1] + step_size

            # Appending new values to solution vectors
            x_vals.append(x_new)
            y_vals.append(y_new)
            y_deriv_vals.append(y_deriv_new)
            y_deriv_deriv_vals.append(y_deriv_deriv_new)

        return y_deriv_deriv_vals, y_deriv_vals, y_vals, x_vals


'''Function for solving a differential equation numerically using modified euler's method
up to 2nd Order.
'''


def mod_euler(f, ics, step_size, x1):
    ''' 
    Function for solving a differential equation numerically using modified euler's method
    up to 2nd Order.
      
    Input variables:
    f :

    - If a 1st order differential equation is required to solve,
                  f is differential equation of the form :
                   dy/dx = f(y, x)

    - If a 2nd order ODE is required to be solved then f = du/dx(u, y, x) where
        u is dy/dx, i.e the first order derivative of the dependent variable with respect to the
        independent variable.

    ics: Initial conditions which should contain at least 2(including x0, which is compulsory) of:
        - x0 = starting value of independent variable (Example: time)
        - y0 = initial value of dependent variable (Example: position)
        - y'0 = initial value of the first derivative of the dependent variable (Example: velocity)
        The order should be (independent variable, dependent variable, 1st derivative of dependent variable,
        second derivative of dependent variable).
        Examples : ics =
                       - 2nd order ODE: [initial time, initial amplitude, initial velocity]
                       - 1st order ODE: [initial time, initial amplitude]

    step_size = value of the 'h' parameter in Euler's method

    x1 = final value of independent variable

    **Returns:**

    - y_deriv_vals = solution vector of y'(x) if a 2nd order ODE is being solved
    - y_vals = solution vector of dependent variable.
    - x_vals = solution vector of the independent variable for which the
    dependent variable was calculated
    '''

    # For solving 1st order ODE
    if len(ics) == 2:

        x0 = ics[0]  # Initial value of independent variable
        y0 = ics[1]  # Initial value of dependent variable

        y_vals = [y0]  # Solution vector for the dependent variable
        x_vals = [x0]  # Solution vector for independent variable
        iterations = int(round((x1 - x0) / step_size))

        '''Number of almost equally spaced points for
         the given step_size and range(x1-x0).
         Iterations are rounded to be able to be used as a range'''

        for i in range(iterations):
            c1 = step_size * f(y_vals[-1], x_vals[-1])  # c1 = h*f(y, x)
            c2 = step_size * f(y_vals[-1] + c1, x_vals[-1] + step_size)  # c2 = h*f(y+c1, x+h)

            next_approx = y_vals[-1] + (0.5 * (c1 + c2))
            '''last element of y vector gets updates as 
                y(x+h) = y(x) + 1/2(c1 + c2)
                c1 and c2 are as defined before '''

            y_vals.append(next_approx)  # Append values of dependent variable to solution vector
            x_vals.append(x_vals[-1] + step_size)  # Append values of independent variable to solution vector

        return y_vals, x_vals

    # For solving 2nd order ODE
    if len(ics) == 3:

        x0 = ics[0]  # Initial value of independent variable
        y0 = ics[1]  # Initial value of dependent variable
        y_deriv0 = ics[2]  # Initial value of first derivative of dependent variable, i.e, y'(y, x)

        y_vals = [y0]  # Solution vector for dependent variable
        y_deriv_vals = [y_deriv0]  # Solution vector for y'(y, x)
        x_vals = [x0]  # Solution vector for independent variable

        iterations = int(round((x1 - x0) / step_size))
        '''Number of almost equally spaced points for
         the given step_size and range(x1-x0).
         Iterations are rounded to be able to be used as a range'''

        for i in range(iterations):
            ''' 
            Here, we shall break up the ODE into 2 ODEs which would form a coupled system of equations.
            - dydx = z(y, x), where z is a function of y and x.
            - dz/dz = f(z, y, x) where f is the given function.
            '''

            # Value of 'C1' for a coupled system
            c1 = step_size * f(y_deriv_vals[-1], y_vals[-1], x_vals[-1])
            m1 = step_size * y_deriv_vals[-1]

            # Value of 'C2' for a coupled system
            m2 = step_size * (y_deriv_vals[-1] + c1)
            c2 = step_size * f(y_deriv_vals[-1] + c1, y_vals[-1] + m1, x_vals[-1] + step_size)

            y_new = y_vals[-1] + (
                    (1 / 2) * (m1 + m2))  # New values of y according to formula y_{n+1} = y_n + 1/2(m1+m2)

            # New values of y according to formula y'_{n+1} = y'_n + 1/2(c1+c2)
            y_deriv_new = y_deriv_vals[-1] + ((1 / 2) * (c1 + c2))

            x_new = x_vals[-1] + step_size  # New values of x as x_{n+1} = x_n + h

            y_vals.append(y_new)  # Append values of dependent variable to solution vector
            y_deriv_vals.append(y_deriv_new)  # Append values of y'(y, x) to solution vector
            x_vals.append(x_new)  # Append values of independent variable to solution vector

        return y_deriv_vals, y_vals, x_vals


''' Range Kutta 4 ODE solver for solving ODEs upto 2nd order '''


def RK4_solver(f, ics, step_size, x1):
    ''' 
    Range Kutta 4 ODE solver for solving ODEs upto 2nd order

    **Input variables:**
    f :

    - If a 1st order differential equation is required to solve,
                  f is differential equation of the form :
                   dy/dx = f(y, x)

    - If a 2nd order ODE is required to be solved then f = du/dx(u, y, x) where
        u is dy/dx, i.e the first order derivative of the dependent variable with respect to the
        independent variable.

    ics: Initial conditions which should contain at least 2(including x0, which is compulsory) of:
        - x0 = starting value of independent variable (Example: time)
        - y0 = initial value of dependent variable (Example: position)
        - y'0 = initial value of the first derivative of the dependent variable (Example: velocity)
        The order should be (independent variable, dependent variable, 1st derivative of dependent variable)

        Example: ics = - 2nd order ODE: [initial time, initial amplitude, initial velocity]
                       - 1st order ODE: [initial time, initial amplitude]

    step_size = value of the 'h' parameter in RK4 method

    x1 = final value of independent variable

    **Returns:**
    - y_deriv_vals = solution vector of y'(x) if a 2nd order ODE is being solved
    - y_vals = solution vector of dependent variable.
    - x_vals = solution vector of the independent variable for which the
    dependent variable was calculated'''

    # For solving first order ODEs
    if len(ics) == 2:

        ''' Formulas used:
        - k1 = hf(y_n, x_n)
        - k2 = hf(y_n + k1/2, x_n + h/2)
        - k3 = hf(y_n + k2/2, x_n + h/2)
        - k4 = hf(y + k3, x + h)

        - y_(n+1) = y_n + 1/6 * (k1 + 2k2 + 2k3 + k4)
        - x_(n+1) = x_n + h
        '''

        x0 = ics[0]  # Initial value of independent variable
        y0 = ics[1]  # Initial value of dependent variable

        y_vals = [y0]  # Solution vector for dependent variable
        x_vals = [x0]  # Solution vector for independent variable

        # Number of iterations based on given step size
        iterations = int(round((x1 - x0) / step_size))
        '''Number of almost equally spaced points for
         the given step_size and range(x1-x0).
         Iterations are rounded to be able to be used as a range'''

        for i in range(iterations):
            k1 = step_size * f(y_vals[-1], x_vals[-1])  # k1 = hf(y_n, x_n)

            # k2 = hf(y_n + k1/2, x_n + h/2)
            k2 = step_size * f(y_vals[-1] + k1 / 2, x_vals[-1] + step_size / 2)

            # k3 = hf(y_n + k2/2, x_n + h/2)
            k3 = step_size * f(y_vals[-1] + k2 / 2, x_vals[-1] + step_size / 2)

            # k4 = hf(y + k3, x + h)
            k4 = step_size * f(y_vals[-1] + k3, x_vals[-1] + step_size)

            # y_(n+1) = y_n + 1/6 * (k1 + 2k2 + 2k3 + k4)
            y_new = y_vals[-1] + 1 / 6 * (k1 + 2 * k2 + 2 * k3 + k4)
            x_new = x_vals[-1] + step_size  # x_(n+1) = x_n + h

            # Solution vector of dependent variable gets updated
            y_vals.append(y_new)
            # Solution vector of independent variable gets updated
            x_vals.append(x_new)

        return y_vals, x_vals  # Returns solution vector of dependent and independent variable

    # For solving 2nd order ODEs

    if len(ics) == 3:

        x0 = ics[0]  # Initial value of independent variable
        y0 = ics[1]  # Initial value of dependent variable
        y_deriv0 = ics[2]  # Initial value of derivative of dependent variable

        # Solution vector for derivative of dependent variable
        y_deriv_vals = [y_deriv0]
        y_vals = [y0]  # Solution vector for dependent variable
        x_vals = [x0]  # Solution vector for independent variable

        # Number of iterations based on given step size
        iterations = int(round((x1 - x0) / step_size))
        '''Number of almost equally spaced points for
         the given step_size and range(x1-x0).
         Iterations are rounded to be able to be used as a range'''

        for i in range(iterations):
            ''' 
            Here, we shall break up the ODE into 2 ODEs which would form a coupled system of equations.
            - dydx = z(y, x), where z is a function of y and x.
            - dz/dz = f(z, y, x) where f is the given function.
            '''

            # Value of 'K1' for the system
            k1 = step_size * f(y_deriv_vals[-1], y_vals[-1], x_vals[-1])
            m1 = step_size * y_deriv_vals[-1]

            # Value of 'K2' for the system
            m2 = step_size * (y_deriv_vals[-1] + (0.5 * k1))
            k2 = step_size * \
                f(y_deriv_vals[-1] + (0.5 * k1), y_vals[-1] +
                  (0.5 * m1), x_vals[-1] + (0.5 * step_size))

            # Value of 'K3' for the system
            m3 = step_size * (y_deriv_vals[-1] + (0.5 * k2))
            k3 = step_size * \
                f(y_deriv_vals[-1] + (0.5 * k2), y_vals[-1] +
                  (0.5 * m2), x_vals[-1] + (0.5 * step_size))

            # Value of 'K4' for the system
            m4 = step_size * (y_deriv_vals[-1] + k3)
            k4 = step_size * \
                f(y_deriv_vals[-1] + k3, y_vals[-1] +
                  m3, x_vals[-1] + step_size)

            # y gets updated as y_new = y_old + 1/6(m1 + 2m2 + 2m3+ m4)
            y_new = y_vals[-1] + ((1 / 6) * (m1 + 2 * m2 + 2 * m3 + m4))

            # y' gets updated as y'_new = y'_old + 1/6(k1 + 2k2 + 2k3+ k4)
            y_deriv_new = y_deriv_vals[-1] + \
                ((1 / 6) * (k1 + 2 * k2 + 2 * k3 + k4))

            # x gets updated as x_new = x_old + h
            x_new = x_vals[-1] + step_size

            # Append values of y'(y, x), y and x to solution vectors
            y_deriv_vals.append(y_deriv_new)
            y_vals.append(y_new)
            x_vals.append(x_new)

        return y_deriv_vals, y_vals, x_vals


if __name__ == '__main__':
    import numpy as np

    # Solving a 1D ODE
    def dydx(y, x):
        return 1/x
    ics1 = [1, 0]
    # Euler method solutions
    y_euler, x_euler = euler(f=dydx, ics=ics1, step_size=0.001, x1=5)

    # Modified euler method solutions
    y_mod, x_mod = mod_euler(f=dydx, ics=ics1, step_size=0.001, x1=5)

    # RK4 solutions
    y_rk4, x_rk4 = RK4_solver(f=dydx, ics=ics1, step_size=0.001, x1=5)

    # Solving a 2D ODE - Equation of a damped harmonis oscillator

    def dudt(u, x, t):  # u = dxdt
        b = 0.2  # damping parameter
        k = 20  # Spring constant
        return -b*u - k*x
    ics = [0, np.pi/50, 0] # At t=0, x = pi/50, vel = 0
   
    # Euler solutions
    u_euler, x_euler2, t_euler = euler(f = dudt, ics = ics, step_size=0.001, x1 = 15)

    # Modified euler solutions
    u_modeuler, x_modeuler2, t_modeuler = mod_euler(f = dudt, ics = ics, step_size=0.001, x1 = 15)

    # RK4 solutions
    u_rk4, x_rk4_2, t_rk4 = RK4_solver(f = dudt, ics = ics, step_size=0.001, x1 = 15)

    plt.plot(x_mod, y_mod, label= 'mod_euler')
    plt.plot(x_euler, y_euler, label = 'euler')
    # plt.plot(u_euler, t_euler, label = 'euler')
    # plt.plot(u_modeuler, t_modeuler, label = 'modeuler')
    # plt.plot(u_rk4, t_rk4, label = 'rk4')


    plt.legend()
    plt.show()

